#!/usr/bin/env python
#
# __init__.py - GLSL shader program management.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``glsl`` package is the home of the :class:`.GLSLShader` class, a class
which may be used to manage GLSL shader programs. The :class:`.GLSLShader`
class is defined in the :mod:`.program` module, and uses some functions
defined in the :mod:`.parse` module.

See also the :mod:`.arbp` package.
"""
