#!/usr/bin/env python
#
# version.py - FSLeyes version information.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The sole purpose of this module is as a container for the *FSLeyes*
version number. See also the :mod:`fsl.version` module.

.. autosummary::
   :nosignatures:

   __version__
"""


__version__ = '1.0.13'
"""Current version number, as a string. The FSLeyes version number consists
of three numbers, separated by a period, which roughly obeys the Semantic
Versioning conventions (http://semver.org/).
"""
